@@ -4,9 +4,9 @@
 ####  ----------------------------------------------
 ## besselI() - Definition as infinite sum  -- working for "mpfr" numbers, too!
 ## ---------
-bI <- function(x, nu, nterm = 800, expon.scaled=FALSE, log = FALSE,
-               Ceps = if(isNum) 8e-16 else 2^(- x@.Data[[1]]@prec))
-
+besselIs <-
+    function(x, nu, nterm = 800, expon.scaled = FALSE, log = FALSE,
+	     Ceps = if(isNum) 8e-16 else 2^(- x@.Data[[1]]@prec))
 {
     ## Purpose: besselI() primitively
     ## ----------------------------------------------------------------------
@@ -15,12 +15,16 @@ bI <- function(x, nu, nterm = 800, expon.scaled=FALSE, log = FALSE,
     ## Author: Martin Maechler, Date: 21 Oct 2002, 21:59
     if(length(nu) > 1)
         stop(" 'nu' must be scalar (length 1)!")
-    j <- (nterm-1):0 # sum smallest first!
     n <- length(x)
     if(n == 0) return(x)
+    j <- (nterm-1):0 # sum smallest first!
+    if(is.numeric(x) && is(nu, "mpfr")) {
+	x <- mpfr(x, precBits = max(64, .getPrec(nu)))
+	isNum <- FALSE
+    } else
+	isNum <- is.numeric(x) || is.complex(x)
     l.s.j <- outer(j, (x/2), function(X,Y) X*2*log(Y))##-> {nterm x n} matrix
-
-    isNum <- is.numeric(x) || is.complex(x)
+    ##
     ## improve accuracy for lgamma(j+1)  for "mpfr" numbers
     ## -- this is very important [evidence: e.g. bI(10000, 1)]
     if(is(l.s.j, "mpfr"))
@@ -31,37 +35,44 @@ bI <- function(x, nu, nterm = 800, expon.scaled=FALSE, log = FALSE,
     ## for large x, this already overflows to Inf :
     ## s.j <- outer(j, (x^2/4), function(X,Y) Y^X) ##-> {nterm x n} matrix
     ## s.j <- s.j / (gamma(j+1) * gamma(nu+1 + j))  but without overflow :
-    log.s.j <- l.s.j - lgamma(j+1) - lgamma(nu+1 + j)
-    if(expon.scaled || log) {
-        ## Compute log I_nu(x) -- trying to avoid overflow/underflow for large x *OR* large nu
-        ## log(s.j)
-        if(log) stop("not yet implemented")
-        ## e..x <- exp(-x) # subnormal for x > 1024*log(2) ~ 710;
-        ## the expon.scaled case:
-        s.j <- exp(log.s.j - rep(x, each=nterm))
-    }
-    else {
-        s.j <- exp(log.s.j)
+    log.s.j <-
+	if(expon.scaled)
+	    l.s.j - rep(abs(Re(x)), each=nterm) - lgamma(j+1) - lgamma(nu+1 + j)
+	else
+	    l.s.j	 	                - lgamma(j+1) - lgamma(nu+1 + j)
+    if(log) {
+	s.j <- lsum(log.s.j) # == log(sum_{j} exp(log.s.j) ); NB: lsum() works on whole matrix
+	if(any(i0 <- x == 0)) s.j[i0] <- 0
+	if(any(lrgS <- Re(log.s.j[1,]) > Re(log(Ceps) + s.j)))
+	    lapply(x[lrgS], function(x)
+		warning(gettextf(" 'nterm=%d' may be too small for %s", nterm,
+				 paste(format(x), collapse=", ")),
+			domain=NA))
+	nu*log(x/2) + s.j
+    } else { ## !log
+	s.j <- ## log I_nu(x) -- trying to avoid overflow/underflow for large x OR large nu
+	    ## log(s.j) ; e..x <- exp(-x) # subnormal for x > 1024*log(2) ~ 710;
+	    exp(log.s.j)
+	s <- colSums(s.j)
+	if(any(i0 <- x == 0)) s[i0] <- 0
+	if(!all(iFin <- is.finite(s)))
+	    stop(sprintf("infinite s for x=%g", x[!iFin][1]))
+	if(any(lrgS <- Re(s.j[1,]) > Re(Ceps * s)))
+	    lapply(x[lrgS], function(x)
+		warning(gettextf(" 'nterm=%d' may be too small for %s", nterm,
+				 paste(format(x), collapse=", ")),
+			domain=NA))
+	(x/2)^nu * s
     }
+}
 
-    s <- x ##- possibly not numeric but "mpfr"
-    for(i in 1:n) {
-        if(x[i] == 0)
-            s[i] <- 0
-        else {
-            sj <- s.j[,i]
-            ## when overflow both in numerator and denominator, drop these:
-            iFin <- is.finite(sj)
-            if(any(!iFin)) stop(sprintf("infinite s.j for x=%g", x[i]))
-            s[i] <- sum(sj)
-            if(sj[1] != 0) {
-                if(abs(sj[1]) > Ceps * abs(s[i]))
-                    warning(sprintf("bI(x=%g): 'nterm' is probably too small",
-                                    x[i]))
-            }
-        }
-    }
-    (x/2)^nu * s
+## old name [back compatibility]:
+bI <- function(x, nu, nterm = 800, expon.scaled = FALSE, log = FALSE,
+               Ceps = if(isNum) 8e-16 else 2^(- x@.Data[[1]]@prec))
+{
+    .Deprecated("besselIs")
+    isNum <- is.numeric(x) || is.complex(x)
+    besselIs(x, nu, nterm=nterm, expon.scaled=expon.scaled, log=log, Ceps=Ceps)
 }
 
 ###--------------- besselI() for large x or also large nu ---
@@ -94,8 +105,8 @@ besselIasym <- function(x, nu, k.max = 10, expon.scaled=FALSE, log=FALSE)
     ## f <- 1
     d <- 0 ## d = 1 - f  <==>  f = 1 - d
     if(k.max >= 1) {
-        m. <- 4*nu^2
-        x8 <- 8*x
+	## m. <- 4*nu^2
+	x8 <- 8*x
         for(k in k.max:1) {
             ## mu <- 4*nu^2; d <- (1 - d)*(mu - (2*k-1)^2)/(k*x8)
             ## mu - (2k-1)^2 = (2nu - (2k-1)) (2nu + (2k-1)) =
@@ -103,11 +114,59 @@ besselIasym <- function(x, nu, k.max = 10, expon.scaled=FALSE, log=FALSE)
             d <- (1 - d)*((2*(nu-k)+1)*(2*(nu+k)-1))/(k*x8)
         }
     }
+    if(expon.scaled)
+        sx <- x -  abs(if(is.complex(x)) Re(x) else x)
+    pi2 <- 2* (if(inherits(x, "mpfr")) Rmpfr::Const("pi", max(.getPrec(x)))
+               else pi)
     if(log) {
         ## f = 1 - d  ==> log(f) = log1p(-d) :
-        (if(expon.scaled) log1p(-d) else x + log1p(-d)) - log(2*pi*x) / 2
+           (if(expon.scaled) sx else x) + log1p(-d) - log(pi2*x) / 2
     } else {
-        (if(expon.scaled) (1-d) else exp(x)*(1-d)) / sqrt(2*pi*x)
+        exp(if(expon.scaled) sx else x) *   (1-d)  / sqrt(pi2*x)
+    }
+}
+
+besselKasym <- function(x, nu, k.max = 10, expon.scaled=FALSE, log=FALSE)
+{
+    ## Purpose: Asymptotic expansion of Bessel K_nu(x) function   x -> oo
+    ##        by Abramowitz & Stegun (9.7.2), p.378 :
+    ##
+    ## K_nu(z) = exp(-z) * sqrt(pi/(2*z)) * f(z,..)  where
+    ##   f(z,..) = 1 + (mu-1)/ (8*z) + (mu-1)(mu-9)/(2! (8z)^2) + ...
+    ##           = 1 + (mu-1)/(8z)*(1 + (mu-9)/(2(8z))*(1 + (mu-25)/(3(8z))*..))
+    ## where  mu = 4*nu^2  *and*  |arg(z)| < pi/2
+    ## ----------------------------------------------------------------------
+    ## Arguments: x, nu, expon.scaled:  as besselK()
+    ## ----------------------------------------------------------------------
+    ## Author: Martin Maechler, Date: 15. Dec 2018
+
+    ##__NB: This *very* similar to besselIasym():  1) different initial factor; 2) *not* alternating
+
+    ## Note that for each x the series eventually *DIVERGES*
+    ## it should be "stopped" as soon as it *converged* (:-)
+
+    stopifnot(k.max == round(k.max))
+
+    ## K_\nu(z) = exp(-z) * sqrt(pi/(2*z)) * f(z, nu)
+
+    ## First compute  f(x, nu) to order k.max
+    ## f <- 1 -- via  d := f - 1  <==>  f = 1 + d
+    d <- 0
+    if(k.max >= 1) {
+	x8 <- 8*x
+        for(k in k.max:1) {
+            ## mu <- 4*nu^2; d <- (1 + d)*(mu - (2*k-1)^2)/(k*x8)
+            ## mu - (2k-1)^2 = (2nu - (2k-1)) (2nu + (2k-1)) =
+            ##               = (2(nu-k)+1)(2(nu+k)-1)
+            d <- (1 + d)*((2*(nu-k)+1)*(2*(nu+k)-1))/(k*x8)
+        }
+    }
+    pi.2 <- (if(inherits(x, "mpfr")) Rmpfr::Const("pi", max(.getPrec(x))) else pi)/2
+    if(log) {
+        ## f = 1 + d  ==> log(f) = log1p(d) :
+        (if(expon.scaled) 0 else   -x   ) + log1p(d)  + (log(pi.2) - log(x)) / 2
+    } else {
+        (if(expon.scaled) 1 else exp(-x)) *  (1+d)    * sqrt(pi.2/x)
     }
 }
 
@@ -119,12 +178,12 @@ besselI.ftrms <- function(x, nu, K = 20)
     ##           = 1- (mu-1)/(8x)*(1- (mu-9)/(2(8x))*(1-(mu-25)/(3(8x))*..))
     ## where  mu = 4*nu^2
     ## ----------------------------------------------------------------------
-    ## Arguments: x, nu, expon.scaled:  as besselI()
+    ## Arguments: x, nu:  as besselI()
     ## ----------------------------------------------------------------------
     ## Author: Martin Maechler, Date: 11 Apr, 21 Nov 2008
 
-    stopifnot(length(x) == 1, length(nu) == 1, length(K) == 1,
-              x > 0, K == round(K), K >= 0)
+    stopifnot(length(x) == 1, x > 0, length(nu) == 1,
+              length(K) == 1, K == round(K), K >= 0)
 
     ## I_\nu(x) = exp(x) / sqrt(2*pi*x) * f(x, nu)
 
@@ -132,18 +191,20 @@ besselI.ftrms <- function(x, nu, K = 20)
     mu <- 4*nu^2
     x8 <- 8*x
     multf <- - (mu - (2*kk-1)^2)/(kk*x8)
-    ser <- cumprod(multf)
 
-    ## now, for k-term approx. f_k  of  f(x,nu)   we have
-    ##  f_k = 1 + sum(ser[1:k]), i.e.
-    ##  (f_k)_k = 1 + cumsum(c(0, ser))[k+1]  for k= 0,1,...,K
-    ser
+    ## ser <- cumprod(multf)
+    ## now, for k-term approximation f_k  of  f(x,nu)   we have
+    ##  f_k = 1 + sum(ser[1:k]), i.e.    ##  (f_k)_k = 1 + cumsum(c(0, ser))[k+1]  for k= 0,1,...,K
+
+    ## ser :=
+    cumprod(multf)
 }
 
 
 ###----------------------------------------------------------------------------
 
 ### When  BOTH  nu and x  are large :
+##        ^^^^ (in practice, works fine already in some cases of |x| large)
 
 besselI.nuAsym <- function(x, nu, k.max, expon.scaled=FALSE, log=FALSE)
 {
@@ -154,17 +215,18 @@ besselI.nuAsym <- function(x, nu, k.max, expon.scaled=FALSE, log=FALSE)
     ##
     ## I_nu(nu * z) ~ 1/sqrt(2*pi*nu) * exp(nu*eta)/(1+z^2)^(1/4) *
     ##                * {1 + u_1(t)/nu + u_2(t)/nu^2 + ... }
-
-    ## where   t = 1 / sqrt(1 + z^2),
-    ##       eta = sqrt(1 + z^2) + log(z / (1 + sqrt(1+z^2)))
-    ##
+    ## where
+    ## __ 9.7.11 __
+    ##	       t := 1 / sqrt(1 + z^2)  = 1/sz
+    ##       eta := sqrt(1 + z^2) + log(z / (1 + sqrt(1+z^2))) = sz + log(z / (1 + sz))
+    ## with   sz := sqrt(1 + z^2)
     ##
     ## and u_k(t)  from  p.366  __ 9.3.9 __
-
+    ##
     ## u0(t) = 1
     ## u1(t) = (3*t - 5*t^3)/24
     ## u2(t) = (81*t^2 - 462*t^4 + 385*t^6)/1152
-    ## ...
+    ## ... up to  u4(t)
 
     ## with recursion  9.3.10    for  k = 0, 1, .... :
     ##
@@ -177,22 +239,30 @@ besselI.nuAsym <- function(x, nu, k.max, expon.scaled=FALSE, log=FALSE)
     ## Author: Martin Maechler, Date: 22 Nov 2008, 15:22
 
     stopifnot(k.max == round(k.max),
-              0 <= k.max, k.max <= 4)
+              0 <= k.max, k.max <= 5)
 
     z <- x/nu # -> computing I_nu(z * nu)
     sz <- sqrt(1 + z^2) ## << "FIXME": use hypot(.)
     t <- 1/sz
-    eta <- (if(expon.scaled) ## <==> * exp(-x)  ==  exp(- z * nu)  <==> "eta - z"
-            ## sz - z = sqrt(1 + z^2) - z  =  1/(sqrt(..) + z)
-            1/(z + sz) else sz) +
-                log(z / (1 + sz))
-
+    ## if(expon.scaled) ## scale by   * exp(-abs(Re(x))):
+    ##     sx <- x -  abs(if(is.complex(x)) Re(x) else x)
+    eta <- (if(expon.scaled) { ## <==> * exp(-|Re(x)|) ==  exp(- |Re(z) * nu|)  <==> "eta - |Re(z)|"
+		## For real z, have
+		##  sz - |z| = sqrt(1 + z^2) - |z|  =!=  1/(sqrt(..) + |z|) = 1/(sz + |z|);
+		## in complex case sz - |x| =  w/(sz + |x|)  with w := 1-y^2 + 2xy i; z = x + iy
+		w <- if(is.complex(z)) {
+			 x. <- Re(z); y <- Im(z)
+			 1-y^2 + 2i*x.*y # complex(real = 1-y^2, imaginary = 2*x.*y)
+		     } else 1
+		w/(sz + abs(if(is.complex(z)) x. else z))
+	    }
+            else sz) + log(z / (1 + sz))
     ## I_nu(nu * z) ~ 1/sqrt(2*pi*nu) * exp(nu*eta)/(1+z^2)^(1/4) *
     ##                * {1 + u_1(t)/nu + u_2(t)/nu^2 + ... }
     if(k.max == 0)
         d <- 0
-    else { ## k.max >= 1 --- Find the  Debye  polynomials u_j(t) in
-           ##                /usr/local/app/R/R_local/src/QRMlib/src/bessel.c
+    else { ## k.max >= 1 --- Find the  Debye  polynomials u_j(t) in ../misc/QRMlib_src_bessel.c
+           ## or in a newer GSL, but much more hidden ..../gsl-2.5/specfunc/debye.c
         t2 <- t^2
         u1.t <- (t*(3 - 5*t2))/24
         d <-
@@ -210,21 +280,36 @@ besselI.nuAsym <- function(x, nu, k.max, expon.scaled=FALSE, log=FALSE)
                     if(k.max == 3)
                         (u1.t + (u2.t + u3.t/nu)/nu)/nu
                     else { ## k.max >= 4
-                        u4.t <- t2*t2*(4465125 +
-                                       t2*(-94121676 +
-                                           t2*(349922430 +
-                                               t2*(-446185740 + t2*185910725))))/39813120
-                    if(k.max == 4)
-                        (u1.t + (u2.t + (u3.t + u4.t/nu)/nu)/nu)/nu
+                        t4 <- t2*t2
+			u4.t <- t4*(4465125 +
+                                    t2*(-94121676 +
+                                        t2*(349922430 +
+                                            t2*(-446185740 + t2*185910725))))/39813120
+			if(k.max == 4)
+			    (u1.t + (u2.t + (u3.t + u4.t/nu)/nu)/nu)/nu
+                        else { ## k.max >= 5
+                            u5.t <- t*t4*(1519035525 +
+                                          t2*(-49286948607 +
+                                              t2*(284499769554 +
+                                                  t2*(-614135872350 +
+                                                      t2*(566098157625 - t2*188699385875))
+                                              )))/6688604160
+                            if(k.max == 5)
+                                (u1.t + (u2.t + (u3.t + (u4.t + u5.t/nu)/nu)/nu)/nu)/nu
+                            else
+                                stop("k.max > 5: not yet implemented (but should NOT happen)")
+                        }
                     }
                 }
             }
     }
 
+    pi2 <- 2* (if(inherits(x, "mpfr")) Rmpfr::Const("pi", max(.getPrec(x)))
+               else pi)
     if(log) {
-        log1p(d) + nu*eta - (log(sz) + log(2*pi*nu))/2
+        log1p(d) + nu*eta - (log(sz) + log(pi2*nu))/2
     } else {
-        (1+d) * exp(nu*eta)/sqrt(2*pi*nu*sz)
+        (1+d) * exp(nu*eta) / sqrt(pi2*nu*sz)
     }
 } ## {besselI.nuAsym}
 
@@ -242,10 +327,10 @@ besselK.nuAsym <- function(x, nu, k.max, expon.scaled=FALSE, log=FALSE)
     ## ----------------------------------------------------------------------
     ## Arguments: x, nu, expon.scaled:  as besselK()
     ## ----------------------------------------------------------------------
-    ## Author: Martin Maechler, Date: 23 Nov 2009, 13:39
+    ## Author: Martin Maechler, Date: 23 Nov 2009
 
     stopifnot(k.max == round(k.max),
-              0 <= k.max, k.max <= 4)
+              0 <= k.max, k.max <= 5)
 
     z <- x/nu # -> computing K_nu(z * nu)
     sz <- sqrt(1 + z^2) ## << "FIXME": use hypot(.)
@@ -258,10 +343,10 @@ besselK.nuAsym <- function(x, nu, k.max, expon.scaled=FALSE, log=FALSE)
     ## K_nu(nu * z) ~  [see above]
     if(k.max == 0)
         d <- 0
-    else { ## k.max >= 1 --- Find the  Debye  polynomials u_j(t) in
-           ##                /usr/local/app/R/R_local/src/QRMlib/src/bessel.c
+    else { ## k.max >= 1 --- Find the  Debye  polynomials u_j(t) ..... [see above!]
         t2 <- t^2
         u1.t <- (t*(3 - 5*t2))/24
+        ## NB: Difference here for K() to I() above is *alternating signs*: "-" for odd uj()
         d <-
             if(k.max == 1) {
                 - u1.t/nu
@@ -277,12 +362,25 @@ besselK.nuAsym <- function(x, nu, k.max, expon.scaled=FALSE, log=FALSE)
                     if(k.max == 3)
                         (- u1.t + (u2.t - u3.t/nu)/nu)/nu
                     else { ## k.max >= 4
-                        u4.t <- t2*t2*(4465125 +
-                                       t2*(-94121676 +
-                                           t2*(349922430 +
-                                               t2*(-446185740 + t2*185910725))))/39813120
-                    if(k.max == 4)
-                        (- u1.t + (u2.t + (-u3.t + u4.t/nu)/nu)/nu)/nu
+                        t4 <- t2*t2
+                        u4.t <- t4*(4465125 +
+                                    t2*(-94121676 +
+                                        t2*(349922430 +
+                                            t2*(-446185740 + t2*185910725))))/39813120
+                        if(k.max == 4)
+                            (- u1.t + (u2.t + (-u3.t + u4.t/nu)/nu)/nu)/nu
+                        else { ## k.max >= 5
+                            u5.t <- t*t4*(1519035525 +
+                                          t2*(-49286948607 +
+                                              t2*(284499769554 +
+                                                  t2*(-614135872350 +
+                                                      t2*(566098157625 - t2*188699385875))
+                                              )))/6688604160
+                            if(k.max == 5)
+                                (- u1.t + (u2.t + (-u3.t + (u4.t - u5.t/nu)/nu)/nu)/nu)/nu
+                            else
+                                stop("k.max > 5: not yet implemented (but should NOT happen)")
+                        }
                     }
                 }
             }
